from decimal import Decimal
from enum import Enum, auto
from typing import Optional

from otlpy.base.order import ORDER_SIDE, Order


class INVENTORY_TYPE(Enum):
    SPOT = auto()
    FUTURE = auto()
    CALL = auto()
    PUT = auto()


class Inventory:
    def __init__(
        self,
        name: str,
        ticker: str,
        /,
        ticksize: Decimal,
        qty_unit: Decimal,
        contract_unit: Decimal,
        fee: Decimal,
        fee_rate: Decimal,
        cash: Decimal,
        pos: Decimal,
        price: Decimal,
        realized_pnl: Decimal,
        realized_fee: Decimal,
        *,
        itype: INVENTORY_TYPE,
        cash_leverage: Decimal,
    ) -> None:
        assert (
            name
            and ticker
            and ticksize > 0
            and qty_unit > 0
            and contract_unit > 0
            and fee >= 0
            and fee_rate >= 0
            and cash_leverage > 0
        )
        self.__name = name
        self.__ticker = ticker
        self.__ticksize = ticksize
        self.__qty_unit = qty_unit
        self.__contract_unit = contract_unit
        self.__fee = fee
        self.__fee_rate = fee_rate
        self.__itype = itype
        self.__cash_leverage = cash_leverage
        self.__cash = cash
        self.__pos = pos
        self.__price = price
        self.__realized_pnl = realized_pnl
        self.__realized_fee = realized_fee
        self.__orders: dict[str, Order] = {}
        self.__opened_buy = Decimal()
        self.__opened_sell = Decimal()

    @property
    def name(self, /) -> str:
        return self.__name

    @property
    def ticker(self, /) -> str:
        return self.__ticker

    @property
    def ticksize(self, /) -> Decimal:
        return self.__ticksize

    @property
    def qty_unit(self, /) -> Decimal:
        return self.__qty_unit

    @property
    def contract_unit(self, /) -> Decimal:
        return self.__contract_unit

    @property
    def fee(self, /) -> Decimal:
        return self.__fee

    @property
    def fee_rate(self, /) -> Decimal:
        return self.__fee_rate

    @property
    def itype(self, /) -> INVENTORY_TYPE:
        return self.__itype

    @property
    def cash_leverage(self, /) -> Decimal:
        return self.__cash_leverage

    @property
    def is_future(self, /) -> bool:
        return self.itype == INVENTORY_TYPE.FUTURE

    @property
    def cash(self, /) -> Decimal:
        return self.__cash

    def _add_cash(self, cash: Decimal, /) -> None:
        self.__cash += cash

    @property
    def pos(self, /) -> Decimal:
        return self.__pos

    def _add_pos(self, pos: Decimal, /) -> None:
        self.__pos += pos

    @property
    def price(self, /) -> Decimal:
        return self.__price

    def _set_price(self, price: Decimal, /) -> None:
        self.__price = price

    @property
    def realized_pnl(self, /) -> Decimal:
        return self.__realized_pnl

    def _add_realized_pnl(self, realized_pnl: Decimal, /) -> None:
        self.__realized_pnl += realized_pnl

    @property
    def realized_fee(self, /) -> Decimal:
        return self.__realized_fee

    def _add_realized_fee(self, realized_fee: Decimal, /) -> None:
        self.__realized_fee += realized_fee

    def init(
        self,
        /,
        *,
        cash: Decimal,
        pos: Decimal,
        price: Decimal,
        realized_pnl: Decimal,
        realized_fee: Decimal,
    ) -> None:
        self.__cash = cash
        self.__pos = pos
        self.__price = price
        self.__realized_pnl = realized_pnl
        self.__realized_fee = realized_fee
        self.clear_orders()

    def unrealized_pnl(self, price: Decimal, /) -> Decimal:
        return (price - self.price) * self.pos * self.contract_unit

    def total_pnl(self, price: Decimal, /) -> Decimal:
        return (
            self.realized_pnl - self.realized_fee + self.unrealized_pnl(price)
        )

    @property
    def opened_buy(self, /) -> Decimal:
        return self.__opened_buy

    def _add_opened_buy(self, opened_buy: Decimal, /) -> None:
        self.__opened_buy += opened_buy

    @property
    def opened_sell(self, /) -> Decimal:
        return self.__opened_sell

    def _add_opened_sell(self, opened_sell: Decimal, /) -> None:
        self.__opened_sell += opened_sell

    @property
    def orders(self, /) -> dict[str, Order]:
        return self.__orders

    @property
    def first_order(self, /) -> Optional[Order]:
        values = list(self.orders.values())
        if values:
            return values[0]
        return None

    def clear_orders(self, /) -> None:
        self.__orders = {}
        self.__opened_buy = Decimal()
        self.__opened_sell = Decimal()

    def add_order(self, order: Order, /, *, mid: Decimal) -> None:
        uid = order.uid
        assert uid and order.opened > 0
        assert self.ticker == order.ticker
        assert self.orders.get(uid) is None
        self.__orders[uid] = order
        if order.oside == ORDER_SIDE.BUY:
            self._add_opened_buy(order.opened)
        elif order.oside == ORDER_SIDE.SELL:
            self._add_opened_sell(order.opened)
        else:
            assert False
        self.logger_info(order, "", mid=mid)

    def filled_position(
        self,
        pos: Decimal,
        price: Decimal,
        /,
    ) -> None:
        s_pos = self.pos
        s_price = self.price
        s_unit = self.contract_unit
        if s_pos * pos > 0:
            realized_pnl = Decimal()
            new_price = (s_price * s_pos + price * pos) / (s_pos + pos)
            self._set_price(new_price)
            self._add_pos(pos)
        elif abs(pos) <= abs(s_pos):
            realized_pnl = (s_price - price) * pos * s_unit
            self._add_realized_pnl(realized_pnl)
            self._add_pos(pos)
        else:
            realized_pnl = (price - s_price) * s_pos * s_unit
            self._add_realized_pnl(realized_pnl)
            self._set_price(price)
            self._add_pos(pos)
        realized_fee = (self.fee + price * s_unit * self.fee_rate) * abs(pos)
        self._add_realized_fee(realized_fee)
        if not self.is_future:
            self._add_cash(-price * pos * s_unit - realized_fee)
        else:
            self._add_cash(realized_pnl - realized_fee)

    def filled_total(
        self,
        order: Order,
        total_filled: Decimal,
        total_filled_price: Decimal,
        total_opened: Decimal,
        /,
        *,
        mid: Decimal,
    ) -> None:
        filled, filled_price, opened = order.filled_total(
            total_filled,
            total_filled_price,
            total_opened,
        )
        if order.oside == ORDER_SIDE.BUY:
            self._add_opened_buy(-opened)
            pos = filled
        elif order.oside == ORDER_SIDE.SELL:
            self._add_opened_sell(-opened)
            pos = -filled
        else:
            assert False
        if pos != 0:
            self.filled_position(pos, filled_price)
        if total_opened == 0:
            del self.__orders[order.uid]
        if filled > 0 or opened > 0:
            self.logger_info(
                order,
                f"\n, FIL, {order.uid}, {filled:f}, {filled_price:f}, {opened:f}",
                mid=mid,
            )

    def logger_info(self, order: Order, /, msg: str, *, mid: Decimal) -> None:
        n = len(self.orders)
        rpnl1 = self.realized_pnl
        rpnl2 = self.realized_fee
        upnl = self.unrealized_pnl(mid)
        tpnl = rpnl1 - rpnl2 + upnl
        order.logger_info(
            f"{msg}"
            f"\n, INV, {self.name}, {n}, {mid:f}"
            f", {tpnl:.2f} = {rpnl1:.2f} - {rpnl2:.2f} + {upnl:.2f}"
            f", {self.pos:f}, {self.price:.2f}"
            f", {self.opened_buy:f}, {self.opened_sell:f}"
        )


class InventoryMap:
    def __init__(self, /) -> None:
        self.__ois: dict[str, tuple[Order, Inventory]] = {}

    @property
    def ois(self, /) -> dict[str, tuple[Order, Inventory]]:
        return self.__ois

    def clear_orders(self, /) -> None:
        for _order, inventory in self.ois.values():
            inventory.clear_orders()
        self.__ois = {}

    @property
    def first_order(self, /) -> Optional[Order]:
        values = list(self.ois.values())
        if values:
            return values[0][0]
        return None

    def add_opened(
        self, order: Order, inventory: Inventory, /, *, mid: Decimal
    ) -> None:
        uid = order.uid
        assert uid and order.opened > 0
        assert order.ticker == inventory.ticker
        assert self.ois.get(uid) is None
        self.__ois[uid] = (order, inventory)
        inventory.add_order(order, mid=mid)

    def filled_total(
        self,
        uid: str,
        total_filled: Decimal,
        total_filled_price: Decimal,
        total_opened: Decimal,
        /,
        *,
        mid: Decimal,
    ) -> None:
        assert uid
        oi = self.ois.get(uid)
        if oi is None:
            return
        order, inventory = oi
        inventory.filled_total(
            order, total_filled, total_filled_price, total_opened, mid=mid
        )
        if total_opened == 0:
            del self.__ois[uid]
